// "Cleansing the Complex", a WASM-4 game
// written by: GNU Radio Shows
// Also published at github.com/gnuradioshows/cg10

// Rotate stick with the arrow keys
// press R to restart the game

#include <math.h>
#include <stdbool.h>

#include "wasm4.h"

static float abs(float x) {
    return x > 0 ? x : -x;
}

static float favor(float a, float d) {
    return d + ((d > 0) == (a > 0) ? a : 0);
}

static bool aligned(float x1, float y1, float x2, float y2, float t) {
    float dot_xy = x1 * x2 + y1 * y2;
    float norm1 = x1 * x1 + y1 * y1;
    float norm2 = x2 * x2 + y2 * y2;
    return dot_xy * dot_xy > t * norm1 * norm2;
}

static bool point_in_rrect(float ox, float oy,
                           float rx, float ry,
                           float r1, float r2) {
    return abs(oy * ry + ox * rx) < r1
        && abs(oy * rx - ox * ry) < r2;
}

static void pixel(uint32_t x, uint32_t y, uint8_t col) {
    uint8_t *bp = &FRAMEBUFFER[(y * 160 + x) >> 2];
    uint8_t shift = x % 4 * 2;
    uint8_t mask = 3 << shift;
    *bp = (*bp & ~mask) | (col << shift);
}

#ifdef DEBUG
int px_checked;
int walls_checked;
int walls_checked_physics;
#endif

static void rrect(float cx, float cy,
                  float rx, float ry,
                  float r1, float r2,
                  uint8_t col)
{
    float rmax = sqrt(r1 * r1 + r2 * r2);

    int xmin = cx - rmax, xmax = cx + rmax + 1;
    if (xmin < 0) xmin = 0;
    if (xmax > 160 - 1) xmax = 160 - 1;

    int ymin = cy - rmax, ymax = cy + rmax + 1;
    if (ymin < 0) ymin = 0;
    if (ymax > 160 - 1) ymax = 160 - 1;

    bool seen_nonempty_row = false;
    for (int y = ymin; y <= ymax; y++) {
        float oy = y + 0.5 - cy;
        bool seen_nonempty_column_in_this_row = false;
        for (int x = xmin; x <= xmax; x++) {
#ifdef DEBUG
            px_checked++;
#endif
            float ox = x + 0.5 - cx;
            if (point_in_rrect(ox, oy, rx, ry, r1, r2)) {
                pixel(x, y, col);
                seen_nonempty_row = true;
                seen_nonempty_column_in_this_row = true;
            } else if (seen_nonempty_column_in_this_row)
                // all subsequent columns in this row will be empty
                break;
        }
        if (seen_nonempty_row && !seen_nonempty_column_in_this_row)
            // all subsequent rows will be empty
            break;
    }
}

static void init_palette(void) {
    // based on: lospec.com/palette-list/ayy4
    PALETTE[0] = 0x00303b;
    PALETTE[1] = 0xf1f2da;
    PALETTE[2] = 0xff7777;
    PALETTE[3] = 0xffce96;
}

static void darker_palette(uint8_t darken) {
#define DARKEN(VAL) ((VAL) < darken ? 0 : (VAL) - darken)
    uint8_t *chan = (uint8_t *)PALETTE;
    *chan++ = DARKEN(0x3b);
    *chan++ = DARKEN(0x30);
    *chan++ = 0;
    *chan++ = 0;
    *chan++ = DARKEN(0xda);
    *chan++ = DARKEN(0xf2);
    *chan++ = DARKEN(0xf1);
    *chan++ = 0;
    *chan++ = DARKEN(0x77);
    *chan++ = DARKEN(0x77);
    *chan++ = DARKEN(0xff);
    *chan++ = 0;
    *chan++ = DARKEN(0x96);
    *chan++ = DARKEN(0xce);
    *chan++ = DARKEN(0xff);
    *chan++ = 0;
#undef DARKEN
}

float px, py, vx, vy, prx, pry = 1, vr;

#define PADDLE_RAD 10
static void paddle(uint8_t gamepad) {
    // motion
    prx -= pry * vr;
    pry += prx * vr;
    px += vx;
    py += vy;
    vy += .05;
    vr *= 0.9;
    // controls
    if (gamepad & BUTTON_LEFT)  vr -= .01;
    if (gamepad & BUTTON_RIGHT) vr += .01;
    // render
    rrect(80, 80, prx, pry, PADDLE_RAD, 1.5, 2);
    rrect(80, 80, prx, pry, PADDLE_RAD - 1, 2, 1);
}

static void debuginfo(uint8_t gamepad) {
#ifdef DEBUG
    static uint8_t last_frame = 0;
    uint8_t button = gamepad & BUTTON_UP;
    if (button && !last_frame) {
        tracef("checked %d pixels while rendering", px_checked);
        tracef("rendered %d, physics for %d",
               walls_checked,
               walls_checked_physics);
        tracef("norm(prx,pry) = %f", prx * prx + pry * pry);
        tracef("px,py = (%f,%f)", px, py);
    }
    last_frame = button;
    px_checked = 0;
    walls_checked = 0;
    walls_checked_physics = 0;
#else
    (void)gamepad;
#endif
}

#define NWALLS ((int)(sizeof(allwalls)/sizeof(struct wall)))
struct wall {
    uint16_t x, y;
    uint8_t dx;
    int8_t dy;
} allwalls[] = {
    // i'm 90% confident this level is possible to win
    // i wouldn't recommend trying
    {3, 2, 0, 10},
    {3, 12, 11, 0},
    {14, 6, 0, 6},
    {17, 7, 0, 6},
    {17, 7, 0, 6},
    {17, 13, 4, 0},
    {18, 2, 3, 1},
    {20, 10, 1, 0},
    {21, 3, 0, 10},
    {28, 3, 10, 0},
    {33, 9, 0, -6},
    {38, 6, 8, 0},
    {38, 9, 3, 0},
    {40, 0, 6, 6},
    {41, 9, 5, 9},
    {46, 12, 0, -6},
    {46, 18, 6, -3},
};

static uint16_t collide_id = 0;

#define ROCKET 2.5
static uint8_t wall(int id, uint16_t x, uint16_t y, uint8_t dx, int8_t dy) {
#define COORD_SCALE ((float)10)
    float dxf = dx, dyf = dy;
    float ax = x * COORD_SCALE - px,
          ay = y * COORD_SCALE - py;
    float bx = ax + dxf * COORD_SCALE,
          by = ay + dyf * COORD_SCALE;
    float cx = (ax + bx) * .5,
          cy = (ay + by) * .5;

    if (ax > 160) return true; // stop iterating
    if (bx < 0 || (by < -10 && ay < -10)
               || (by > 170 && ay > 170)) return false; // not drawn

#ifdef DEBUG
    walls_checked++;
#endif

    float length = sqrt(dxf * dxf + dyf * dyf);
    float rx = dxf / length;
    float ry = dyf / length;
    float r1 = length * 5 + 1;
    float r2 = 3;
    rrect(cx, cy, rx, ry, r1, r2, 3);

    if (collide_id) return false; // skip all collisions
    if ((ax > 80+15 && bx > 80+15) || (ay > 80+15 && by > 80+15)
     || (ay < 80-15 && by < 80-15) || (ay < 80-15 && by < 80-15))
        return false; // out of bounds to possibly be a collision

#ifdef DEBUG
    walls_checked_physics++;
#endif

    // wall corner pixels
    float co1x = cx + r1 * rx - r2 * ry,
          co1y = cy + r1 * ry + r2 * rx,
          co2x = cx + r1 * rx + r2 * ry,
          co2y = cy + r1 * ry - r2 * rx,
          co3x = cx - r1 * rx - r2 * ry,
          co3y = cy - r1 * ry + r2 * rx,
          co4x = cx - r1 * rx + r2 * ry,
          co4y = cy - r1 * ry - r2 * rx;

    float pr1 = PADDLE_RAD - 4,
          pr2 = 2.0;

    // paddle corner pixels
    float co5x = 80 + pr1 * prx - pr2 * pry,
          co5y = 80 + pr1 * pry + pr2 * prx,
          co6x = 80 + pr1 * prx + pr2 * pry,
          co6y = 80 + pr1 * pry - pr2 * prx,
          co7x = 80 - pr1 * prx - pr2 * pry,
          co7y = 80 - pr1 * pry + pr2 * prx,
          co8x = 80 - pr1 * prx + pr2 * pry,
          co8y = 80 - pr1 * pry - pr2 * prx;

    if (point_in_rrect(80 - co1x, 80 - co1y, prx, pry, pr1, pr2)
     || point_in_rrect(80 - co2x, 80 - co2y, prx, pry, pr1, pr2)
     || point_in_rrect(80 - co3x, 80 - co3y, prx, pry, pr1, pr2)
     || point_in_rrect(80 - co4x, 80 - co4y, prx, pry, pr1, pr2)
    ) {
        // collision between paddle & wall corner
        collide_id = id;
        // apply a normal force which is perpendicular to the paddle
        // and opposite the direction of its velocity
        float nx = -pry;
        float ny = prx;
        if (nx * vx < ny * vy) {
            nx = -nx;
            ny = -ny;
        }
        // yes, i know about the bug where you get stuck in walls here,
        // but for the life of me i haven't figured out how to fix it
        px += vx = nx;
        py += vy = ny;
    } else if (point_in_rrect(co5x - cx, co5y - cy, rx, ry, r1, r2)
            || point_in_rrect(co6x - cx, co6y - cy, rx, ry, r1, r2)
            || point_in_rrect(co7x - cx, co7y - cy, rx, ry, r1, r2)
            || point_in_rrect(co8x - cx, co8y - cy, rx, ry, r1, r2)
    ) {
        // collision between paddle corner & wall
        collide_id = id;
        px += vx = (-0.9 * vx);
        py += vy = (-0.9 * vy);
    } else if (point_in_rrect(
        80 - cx + PADDLE_RAD * prx,
        80 - cy + PADDLE_RAD * pry,
        rx, ry, r1, r2)
    ) {
        // collision between paddle back tip & wall
        collide_id = id;
        vx = favor(vx, -ROCKET * prx);
        vy = favor(vy, -ROCKET * pry);
    } else if (point_in_rrect(
        80 - cx - PADDLE_RAD * prx,
        80 - cy - PADDLE_RAD * pry,
        rx, ry, r1, r2)
    ) {
        // collision between paddle front tip & wall
        collide_id = id;
        vx = favor(vx, ROCKET * prx);
        vy = favor(vy, ROCKET * pry);
    }

    return false; // checked for collision
}

static void walls(void) {
    collide_id = 0;
    for (int i = 0; i < NWALLS; i++) {
        struct wall w = allwalls[i];
        if (wall(i + 1, w.x, w.y, w.dx, w.dy)) break;
    }
    static float ppx, ppy, pprx, ppry;
    if (collide_id) {
        px = ppx;
        py = ppy;
        prx = pprx;
        pry = ppry;
    } else {
        ppx = px;
        ppy = py;
        pprx = prx;
        ppry = pry;
    }
}

void check_game(void) {
    static uint8_t loss_frame = 0;
    static uint32_t frame_count = 0;
    static bool did_win = false;

    if (!did_win)
        frame_count++;

    {
        uint32_t seconds = frame_count / 60;
        uint32_t minutes = seconds / 60;
        static char fmt[6];
        if (minutes > 99) {
            fmt[0] = fmt[1] = 'x';
        } else {
            fmt[0] = minutes > 9 ? '0' + minutes / 10 : ' ';
            fmt[1] = '0' + minutes % 10;
        }
        fmt[2] = ':';
        fmt[3] = '0' + (seconds % 60 / 10);
        fmt[4] = '0' + (seconds % 10);
        fmt[5] = 0;
        *DRAW_COLORS = 0x3;
        text(fmt, 120, 1);
    }

    if ((px - 390) * (px - 390) + (py - 65) * (py - 65) < 100)
        did_win = true;

    if (did_win) {
        *DRAW_COLORS = 0x4;
        text("You win!", 10, 10);
    } else if (loss_frame) {
        if (loss_frame < 30) {
            int darken = loss_frame * 17;
            darker_palette(darken > 255 ? 255 : darken);
        } else if (loss_frame < 90) {
            int red = (loss_frame - 30) * 17;
            PALETTE[3] = (red > 255 ? 255 : red) << 16;
            *DRAW_COLORS = 0x4;
            text("You lose", 10, 10);
        } else if (loss_frame <= 120) {
            frame_count = 0;
            px = py = vx = vy = vr = 0;
            prx = 0; pry = 1;
            int darken = 255 - (loss_frame - 90) * 14;
            darker_palette(darken < 0 ? 0 : darken);
        } else if (loss_frame == 121) {
            loss_frame = 0;
            return;
        }
        loss_frame++;
    } else if (py > 200) {
        loss_frame = 1;
    }
}

void start(void) {
    init_palette();
}

void update(void) {
    check_game();
    uint8_t gamepad = *GAMEPAD1;
    walls();
    paddle(gamepad);
    debuginfo(gamepad);
}
